package directSpread;

import java.awt.Shape;
import java.awt.geom.Rectangle2D;
import java.util.Arrays;
import java.util.Random;

import directSpread.utility.Operator;
import directSpread.utility.Solution;
import directSpread.utility.Variable;
import java.io.Serializable;
import windCode.Area;
import windCode.RectangularArea;

/**
 * Mutates a coordinate by moving it away from its nearest neighbours.
 *  The vectors between the coordinate and its nearest neighbours are combined
 *  to create a displacement vector; the direction of this vector is modified by
 *  a normal distribution and the distance set with a normal distribution based
 *  on a default length.
 *
 *  Operates on a single coordinate per entry.
 */
public class OpenWindMutation implements Operator, Serializable {

    private Random random = new Random();
    private int nearestNeighbours;
    //private double directionSigma;
    //private double distanceMu;
    //private double distanceSigma;
    private Shape bounds;
    private double minimumDistance;
    private double probability;
    private double reversalProbability;
    private Area area;
    private double width;
    private double height;
    private boolean[] mutatedNodes;

//TODO determine failure within the class or without?
//        public void setDistanceMuAndSigma(double mu, double sigma) {
//            this.distanceMu = mu;
//            this.distanceSigma = sigma;
//        }
//
//        public double getDistanceMu() {
//            return distanceMu;
//        }
//
//        public double getDistanceSigma() {
//            return distanceSigma;
//        }
    /**
     * Creates an instance of the Vector-Based Movement Mutation for coordinate systems.
     *
     * @param probability  The probability that this mutation will occur.
     * @param nearestNeighbours  The number of nearest neighbours to use to construct a
     *  displacement vector. This must be greater than 0.
     * @param bounds  The area the nodes may occupy.
     * @param minimumDistance  The minimumDistance between two nodes.
     * @param directionSigma  The standard deviation for the direction.
     * @param distanceMu  The mean distance to move.
     * @param distanceSigma  The standard deviation for the distance.
     * @param reversalProbability  The probability the displacement will be reversed.
     */
    public OpenWindMutation(double probability, int nearestNeighbours, Shape bounds, Area area, double width, double height, double minimumDistance,
            /*double directionSigma, double distanceMu, double distanceSigma,*/ double reversalProbability) {
        if (nearestNeighbours < 0) {
            System.err.println("Error in VectorBasedMovement constructor: nearestNeighbours cannot be < 0.");
            System.exit(1);
        }
        this.probability = probability;
        this.nearestNeighbours = nearestNeighbours;
        this.bounds = bounds;
        this.minimumDistance = minimumDistance;
//		this.directionSigma = directionSigma;
//		this.distanceMu = distanceMu;
//		this.distanceSigma = distanceSigma;
        this.reversalProbability = reversalProbability;
        this.area = area;
        this.width = width;
        this.height = height;
    }

    /**
     * Works on a series of x,y coordinates in the form [x0, y0, x1, y1, ...].
     * Selects and mutates a coordinate by moving it away from its nearest neighbours.
     *
     * @param object The solution to mutate.
     * @return  The mutated solution.
     */
    public Object execute(Object object) {

        Solution solution = (Solution) object;
        mutatedNodes = new boolean[solution.getDecisionVariables().length / 2];
        doMutation(probability, solution);

        return solution;
    }

    public boolean[] getMutatedNodes() {
        return mutatedNodes;
    }
    
    
    Variable vectors[] = null;
    boolean[] bMovers = null;
    boolean[] bShakers = null;
    boolean[] bMoving = null;

    /**
     * Works on a series of x,y coordinates in the form [x0, y0, x1, y1, ...].
     * Selects and mutates a coordinate by moving it away from its nearest neighbours.
     * Current weight for each neighbour vector v is 1/||v||.
     *
     * @param probability  The probability that the mutation will occur.
     * @param solution  The solution to mutate.
     */
    private boolean doMutation(double probability, Solution solution) {
        
        boolean debugPrint = !true;
        
        
        //TODO Look at how the function is being called. Might have to alter the way the mutation works.
        
        
        //m_bStop - TODO unsure
        boolean m_bStop = false;
        
        
        //trbs - the turbines
        //pos - the position of the turbines
        //tempts - perturbed turbine positions (ie, keep originals and mutated set.
        //vectors - the amount moved. This is 0 when entering the function
        //origins - checked for validity - ignore.
    	// bMovers - whether each turbine is moveable or not. Set all to true.
    	// bShakers - whether each turbine's energy is counted in assessing improvement. Set all to true. 
        //bMoving - size = number of turbines. Set all to false.
        
        Variable[] variables = solution.getDecisionVariables();
        int n = solution.getDecisionVariables().length / 2;

        
        //trbs. Not using as stated above.
        Variable[] pos = variables;
        Variable[] tempts = new Variable[2*n];
      //origins. Not using as stated above

        //use resets (line does not come from the original source)
        vectors=null;
        
        //Initiate things.
        if (vectors == null){
	        vectors = new Variable[2*n];
	        bMovers = new boolean[n];
	        bShakers = new boolean[n];
	        bMoving = new boolean[n];
	        
	        for (int i = 0; i < n; i++){
	        	bMovers[i] = true;
	        	bShakers[i]=true;
	        	vectors[2*i] = new Variable(0);
	        	vectors[2*i+1] = new Variable(0);
	        	bMoving[i] = false;
	        }
        }
        
        double f = 1.0;
        
    	//duplicate pos into tempts
        for (int i = 0; i < n; i++){
        	tempts[2*i] = new Variable(pos[2*i].getValue());
        	tempts[2*i+1] = new Variable(pos[2*i+1].getValue());
        }
        
        for (int i = 0; i < n; i++){
    		if(!bMovers[i])
    		{
    			continue; // i.e. dont perturb	
    		}
    		if(bMoving[i])
    		{
    			tempts[2*i].setValue(pos[2*i].getValue()+vectors[2*i].getValue());
    			tempts[2*i+1].setValue(pos[2*i+1].getValue()+vectors[2*i+1].getValue());
    			
    			//C++ version
    			/*if(legal.IsLegal(pMain,trbs,origins,tempts,true) && legal.CheckWRGs(pMain, trbs, origins, tempts))
    				continue;*/	
 
    			//If the altered position is legal, the mutation is done.
		        Neighbour[] temptsNeighbours = computeNeighbours(tempts, i);   			
    			if (isInBounds(tempts[i].getValue(), tempts[i].getValue()) && !isTooCloseToOtherNodes(tempts, i, temptsNeighbours) && !isWithinInfeasibleZone(tempts, i)){
    				
    				continue;
    			}
    		}
    		
    		Neighbour[] temptsNeighbours = null;
    		
    		int c=-100; 
    		do
    		{
    			f = c<=0 ? 1.0 : 1.0/(double)c;			
    			
    			//TODO check if this is what their random function does. is it between 0 and 1, or -1 and 1..?
    			double arg1,arg2,rad = random.nextDouble();
    			rad *= rad;
    			rad = Math.max(rad,0.005);
    			rad *= rad; // instead of squaring below
    			
    			do{
    				arg1 = random.nextDouble();
    				arg2 = random.nextDouble();
    				arg1 = arg1*2.0-1.0;
    			}while(Math.abs(arg2) >= Math.exp(-arg1*arg1/(rad)) && !m_bStop);

    			if(arg1>0.0)
    				vectors[2*i].setValue( arg1*(bounds.getBounds().width-pos[2*i].getValue())*f);
    			else
    				vectors[2*i].setValue(arg1*(pos[2*i].getValue()-bounds.getBounds().x)*f);
    				
    			do{
    				arg1 = random.nextDouble();
    				arg2 = random.nextDouble();
    				arg1 = arg1*2.0-1.0;
    			}while(Math.abs(arg2) >= Math.exp(-arg1*arg1/(rad)) && !m_bStop);
    			// ...to here

    			if(arg1>0.0)
    				vectors[2*i+1].setValue( arg1*(bounds.getBounds().y-pos[2*i+1].getValue())*f);
    			else
    				vectors[2*i+1].setValue( arg1*(pos[2*i+1].getValue()-bounds.getBounds().y+bounds.getBounds().height)*f);
    			
    			tempts[2*i].setValue(pos[2*i].getValue()+vectors[2*i].getValue());
    			tempts[2*i+1].setValue(pos[2*i+1].getValue()+vectors[2*i+1].getValue());
    			
    			bMoving[i]=true;
    			
    			if(m_bStop)
    			{
    				return false;
    			}
    			
    			c++;
    			
    			if(c>100)
    			{
    				vectors[2*i].setValue(0.0);
    				vectors[2*i+1].setValue(0.0);
    				tempts[2*i].setValue(pos[2*i].getValue());
    				tempts[2*i+1].setValue(pos[2*i+1].getValue());
    				bMoving[i] = false;
    				break;
    			}
    			
    			//Don't need to output or display anything.
//    			DUMP(c);
    			
    			temptsNeighbours = computeNeighbours(tempts, i);
    			
    		}while(!isInBounds(tempts[2*i].getValue(), tempts[2*i+1].getValue()) || isTooCloseToOtherNodes(tempts, i, temptsNeighbours) || isWithinInfeasibleZone(tempts, i));
    		
    		// ...to here and then check for conflicts
    		// here...
    		
//    		if(n>300) // may take a while so update user...
    		
			//Don't need to output or display anything.
    		//PostCallback(callback1(ptr,&OpenWind::SetStatus,Format("Perturbing..{%d}.",i+1)));
    		
        }
        
        //If legal, accept changes. Otherwise don't.
        boolean legal = true;
        for (int i = 0; i < n; i++){
            Neighbour[] temptsNeighbours = computeNeighbours(tempts, i);
        	if (!isInBounds(tempts[2*i].getValue(), tempts[2*i+1].getValue()) || isTooCloseToOtherNodes(tempts, i, temptsNeighbours) || isWithinInfeasibleZone(tempts, i)){
        		legal = false;
        		break;
        	}
        }
        if (legal){
	        for (int i = 0; i < n; i++){
	        	if (pos[2*i].getValue() != tempts[2*i].getValue() || pos[2*i+1].getValue() != tempts[2*i+1].getValue()){
	        		mutatedNodes[i] = true;
	        		pos[2*i].setValue(tempts[2*i].getValue());
	        		pos[2*i+1].setValue(tempts[2*i+1].getValue());
		        	solution.turbineMoved[i] = true;
	        	}
	        	
	        }
        }else{
        	for (int i = 0 ; i < n; i++){
        		bMoving[i] = false;
        	}
        }
        
    	return true;
       
        
        
        
        
        
//        
//        if (random.nextDouble() >= probability) {
//            return;
//        }
//
//        Variable[] variables = solution.getDecisionVariables();
//        int numNodes = solution.getDecisionVariables().length / 2;
//        if (nearestNeighbours >= numNodes) {
//            System.err.println("Error in VectorBasedMovement.doMutation: nearestNeighbours must be less than the number of nodes.");
//            System.exit(1);
//        }
//
////	for (int currentNode = 0; currentNode < numNodes; currentNode++){
////		if (random.nextDouble() < probability) continue;
//
//        int currentNode = random.nextInt(numNodes);
//        if (debugPrint) System.out.println("doMutation: "+currentNode+" of "+numNodes);
//        mutatedNodes[currentNode] = true;                                       // needed for sigma-adaptation
//        solution.turbineMoved[currentNode] = true;                              // needed for faster fitness evaluations
//        
//
//        double currentX = variables[2 * currentNode].getValue();
//        double currentY = variables[2 * currentNode + 1].getValue();
//
//        //Get nearest neighbours.
//        Neighbour[] neighbours = computeNeighbours(variables, currentNode);
//
//        //Get displacement vector in x,y form.
//        double displacementX = 0;
//        double displacementY = 0;
//        for (int i = 0; i < nearestNeighbours; i++) {
//            int currentNeighbour = neighbours[i].getNeighbourIndex();
//            double distance = neighbours[i].getDistanceFromNode();
//
//            displacementX += (currentX - variables[2 * currentNeighbour].getValue()) / distance;
//            displacementY += (currentY - variables[2 * currentNeighbour + 1].getValue()) / distance;
//        }
//
//        //Get angle and new distance of displacement vector
//        double angle = getDisplacementDirection(solution, currentNode, displacementX, displacementY);
//        double distance = getDisplacementDistance(solution, currentNode);
//
//        //Determine mutation
//        displacementX = distance * Math.cos(angle);
//        displacementY = distance * Math.sin(angle);
//        int reverseDisplacement = 1;
//        int reduction = 8;
//
//        //Non-deterministic reversal
//        if (random.nextDouble() < reversalProbability) {
//            reverseDisplacement = -reverseDisplacement;
//        }
//
//        //Determine mutation
//        displacementX = distance * Math.cos(angle) * reverseDisplacement;
//        displacementY = distance * Math.sin(angle) * reverseDisplacement;
//
//        double newX = currentX + displacementX;
//        double newY = currentY + displacementY;
//
//
//
//        //Update solution
//        variables[2 * currentNode].setValue(newX);
//        variables[2 * currentNode + 1].setValue(newY);
//
//        //Attempt to force new coordinates into legal position.
//        double originalDistance = distance;
//        neighbours = computeNeighbours(variables, currentNode);
//        while (!isInBounds(newX, newY) || isTooCloseToOtherNodes(variables, currentNode, neighbours) || isWithinInfeasibleZone(variables, currentNode)) {
//            variables[2 * currentNode].setValue(currentX);
//            variables[2 * currentNode + 1].setValue(currentY);
//            distance -= originalDistance / reduction;
//
//            if (distance < 0) {
//                break;
//            }
//
//            displacementX = distance * Math.cos(angle) * reverseDisplacement;
//            displacementY = distance * Math.sin(angle) * reverseDisplacement;
//            newX = currentX + displacementX;
//            newY = currentY + displacementY;
//            variables[2 * currentNode].setValue(newX);
//            variables[2 * currentNode + 1].setValue(newY);
//            neighbours = computeNeighbours(variables, currentNode);
//        }



//	    if (!isInBounds(newX, newY)){
//                if (newX>width) newX=width;
//                if (newY>height) newY=height;
//                if (newX<0) newX=0;
//                if (newY<0) newY=0;
//
//	    }
////            else{
//		    //Update solution
//		    variables[2*currentNode].setValue(newX);
//		    variables[2*currentNode+1].setValue(newY);
//
//		    //Reset solution if too close to other nodes.
//		    neighbours = computeNeighbours(variables, currentNode);
//		    if (isTooCloseToOtherNodes(variables, currentNode, neighbours)){
//	    	    variables[2*currentNode].setValue(currentX);
//	    	    variables[2*currentNode+1].setValue(currentY);
////		    }
//	    }

        //}
    }

    /**
     * Returns whether the coordinates are within the bounding shape or not.
     * @param x  The x coordinate.
     * @param y  The y coordinate.
     * @return  True if the x,y coordinates are within the bounding shape, false otherwise.
     */
    private boolean isInBounds(double x, double y) {
        return bounds.contains(x, y);
    }

    /**
     * Returns whether the specified node is too close to the supplied neighbours.
     * @param variables
     *
     * @param variables  The list of all node coordinates.
     * @param currentNode  The specified node.
     * @param neighbours  All other nodes.
     * @return  True if the specified node is too close to the supplied neighbours, false otherwise.
     */
    private boolean isTooCloseToOtherNodes(Variable[] variables, int currentNode, Neighbour[] neighbours) {
        if (neighbours.length == 0) {
            return false;
        }

        if (determineDistance(variables, currentNode, neighbours[0].getNeighbourIndex()) < minimumDistance) {

//       	if (neighbours[0].getDistanceFromNode() < minimumDistance){
            //Reset variables.
            return true;
        } else {
            return false;
        }
    }

    /**
     * Returns a direction based on the supplied vector. Uses a normal
     *  distribution if the length of the supplied vector is more than 0,
     *  otherwise returns a completely random direction.
     *
     * @param displacementX  The x displacement of the vector.
     * @param displacementY  The y displacement of the vector.
     * @return  The new direction in radians.
     */
    private double getDisplacementDirection(Solution solution, int currentNode, double displacementX, double displacementY) {

        double distance = Math.sqrt(displacementX * displacementX + displacementY * displacementY);
        //If distance is zero, return random direction.
        if (distance == 0) {
            return random.nextDouble() * Math.PI * 2;
        }
//		double sigma = directionSigma;
        double sigma = solution.getDirectionSigma(currentNode);

        //Determine mu based of the normalised X and Y values.
        //Note that acos, asin and atan return values within one pi range.
        double mu = 0;
        if (displacementX < 0 && displacementY < 0) {
            mu = Math.atan(displacementY / displacementX) + Math.PI;
        } else if (displacementY < 0) {
            mu = Math.asin(displacementY / distance);
        } else {
            mu = Math.acos(displacementX / distance);
        }

        //Obtain the angle based on a normal distribution.
        double angle = normalDistribution(mu, sigma) % (2 * Math.PI);
        return angle;
    }

    /**
     * Returns a normally distributed distance based on the global distance
     *  mean and standard deviation.
     *
     * @return  a normally distributed distance based on the global distance
     *  mean and standard deviation.
     */
    private double getDisplacementDistance(Solution solution, int currentNode) {
        //return normalDistribution(distanceMu, distanceSigma);
//		return normalDistribution(solution.getDistanceMu(currentNode), solution.getDistanceSigma(currentNode));
        return Math.abs(normalDistribution(0, solution.getDistanceSigma(currentNode)));

    }

    /**
     * Returns a value based on a normal distribution with the supplied mean mu
     *  and standard deviation sigma.
     *
     * @param mu  The mean for the normal distribution.
     * @param sigma  The standard deviation for the normal distribution.
     * @return  a value based on a normal distribution with the supplied mean mu
     *  and standard deviation sigma.
     */
    private double normalDistribution(double mu, double sigma) {
        return random.nextGaussian() * sigma + mu;
    }

    /**
     * Assumes a [x0, y0, x1, y1, ...] x,y coordinate set for nodes 0, 1, ...
     *
     * @param node1  First node with x at index 2*node1, y at index 2*node1+1
     * @param node2  Second node with x at index 2*node2, y at index 2*node2+1
     * @return The distance between these two nodes. This is set to
     *  Double.MAX_VALUE if the two nodes are the same.
     */
    private double determineDistance(Variable[] variables, int node1, int node2) {

        //
        if (node1 == node2) {
            return Double.MAX_VALUE;
        }

        double xDiff = variables[2 * node2].getValue() - variables[2 * node1].getValue();
        double yDiff = variables[2 * node2 + 1].getValue() - variables[2 * node1 + 1].getValue();

        return Math.sqrt(xDiff * xDiff + yDiff * yDiff);
    }

    /**
     * Creates an ordered array of the nearest neighbours to a single node.
     * The provided node is not included in the array.
     *
     * @param variables  The coordinates for all nodes
     * @param node  The node being used to calculate distances to all other nodes.
     * @return  An array of indices and distances for neighbours to the closest node.
     */
    private Neighbour[] computeNeighbours(Variable[] variables, int node) {

        int numNodes = variables.length / 2;
        Neighbour neighbours[] = new Neighbour[numNodes - 1];

        //Get distance from this node to all others.
        for (int neighbour = 0; neighbour < node; neighbour++) {
            neighbours[neighbour] = new Neighbour(neighbour, determineDistance(variables, node, neighbour));
        }
        for (int neighbour = node + 1; neighbour < numNodes; neighbour++) {
            neighbours[neighbour - 1] = new Neighbour(neighbour, determineDistance(variables, node, neighbour));
        }

        //Return the indices of the nodes ordered by closest.
        Arrays.sort(neighbours);
        return neighbours;
    }

    /**
     * Returns whether the current variable is inside an infeasible zone.
     * @param variables  The coordinates for all nodes.
     * @param currentNode  The current node.
     * @return  true if the current variable is inside an infeasible zone, else false.
     */
    private boolean isWithinInfeasibleZone(Variable[] variables, int currentNode) {
        Rectangle2D[] infeasibleZone = ((RectangularArea) area).getInfeasibleZones();

        double x = variables[2 * currentNode].getValue();
        double y = variables[2 * currentNode + 1].getValue();

        for (int i = 0; i < infeasibleZone.length; i++) {
        	if (infeasibleZone[i] == null) break;
            if (infeasibleZone[i].contains(x, y)) {
                return true;
            }
        }
        return false;
    }

    /**
     * InnerStorage class allowing easy permutation of neighbour indices and distances.
     */
    class Neighbour implements Comparable<Neighbour> {

        private int neighbourIndex;
        private double distanceFromNode;

        /**
         * Storage class constructor; takes in both which index of the node
         *  list this neighbour is and how far away it is from the current node.
         *
         * @param neighbourIndex  The index of the node list this neighbour is.
         * @param distanceFromNode  The distance from this neighbour to the
         *  current node.
         */
        public Neighbour(int neighbourIndex, double distanceFromNode) {
            this.neighbourIndex = neighbourIndex;
            this.distanceFromNode = distanceFromNode;
        }

        /**
         * Compares this Neighbour to the supplied by comparing the
         *  distances.
         *
         *  @param o  The neighbour to compare against this neighbour.
         */
        public int compareTo(Neighbour o) {
            return Double.compare(distanceFromNode, o.distanceFromNode);
        }

        /**
         * Returns the index of the node list this neighbour is.
         * @return  The index of the node list this neighbour is.
         */
        public int getNeighbourIndex() {
            return neighbourIndex;
        }

        /**
         * Returns the distance from this neighbour to the current node.
         * @return  The distance from this neighbour to the current node.
         */
        public double getDistanceFromNode() {
            return distanceFromNode;
        }
    }
}