package directSpread;

import deepCopy.DeepCopy;
import java.awt.Shape;
import java.awt.geom.Rectangle2D;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Map;
import java.util.Random;
import java.util.TreeMap;

import windCode.Area;
import windCode.WakeModel;

import directSpread.utility.Operator;
import directSpread.utility.Problem;
import directSpread.utility.Solution;
import directSpread.utility.SolutionSet;
import directSpread.utility.Variable;
import directSpread.utility.WindTurbineProblemFaster;
import java.io.Serializable;
import windCode.WakeModelFaster;

public class WindTurbineAlgorithmFaster implements Serializable{
    
    

	private Problem problem;
	private int maxEvaluations;
	private int populationSize;
	private WindTurbineGUI gui;
	private Operator mutationOperator;
    private boolean printEveryEvaluation = true;
	private Random random = new Random();
	private boolean printSeveralOutputsPerDSRun;
	private boolean displayGUI;
	
	private boolean pauseForGui = !true; //Used to take screenshots of GUI.

		/**
	 * Instantiate the Wind Turbine algorithm.
	 * 
	 * @param problem  The Problem to operate on.
	 * @param bounds  The bounds of the turbine placement area.
	 * @param minimumDistance  The minimum distance allowed between two turbines.
	 * @param maxEvaluations  The number of evaluations to perform.
	 * @param populationSize  The number of Solutions in the population.
	 */
	public WindTurbineAlgorithmFaster(Problem problem, Shape bounds, double minimumDistance, Operator mutationOperator, int maxEvaluations, int populationSize,
                boolean printEveryEvaluation, boolean printSeveralOutputsPerDSRun, boolean displayGUI, Rectangle2D[] infeasibleZones){
		this.problem = problem;
		this.mutationOperator = mutationOperator;
		this.maxEvaluations = maxEvaluations;
		this.populationSize = populationSize;		
		this.printSeveralOutputsPerDSRun = printSeveralOutputsPerDSRun;
		this.printEveryEvaluation = printEveryEvaluation;
		this.displayGUI = displayGUI;
		
		if (displayGUI) gui = new WindTurbineGUI(minimumDistance, bounds, 600, 600, infeasibleZones);
		
		
		
		
//		
//		
//		Area area, WakeModel fitfun, int mu, int lambda, int generations,
//        boolean printSeveralOutputsPerDSRun, double initialStandardDeviation,
//        int workerThreads
	}
	
	/**
	 * Launches the execution of the Wind Turbine algorithm.
	 * 
	 * @return  The resulting population of turbine coordinates.
	 */
	public SolutionSet execute(){
        boolean debugPrint = !true;
//		Operator mutationOperator;
        
        //some additional printouts and statistics
        WakeModelFaster fitfun = ((WindTurbineProblemFaster)problem).getWakeModel();
        Area area = ((WindTurbineProblemFaster)problem).getArea();
        double[] startingPoint = area.startingSolution();
        Solution startingSolution = new Solution(startingPoint);
        double wakelossFreeFitness = fitfun.computeWakelossFreeFitness(startingPoint);
        
////        System.out.println("initial="+1d/fitfun.valueOf(startingPoint, true)+", wakelossfree="+1d/wakelossFreeFitness);
//        try{
//            System.out.println("initial="+1d/fitfun.valueOf(startingPoint, printEveryEvaluation, null, -1)+", wakelossfree="+1d/wakelossFreeFitness);
////            System.out.println("initial="+1d/fitfun.valueOf(startingPoint, printEveryEvaluation, null, -1)+", wakelossfree="+1d/wakelossFreeFitness);
//        } catch (Exception e) {
//            // don't do anything, the nullpointerexception is harmless
//        }

        
        fitfun.valueOf(true, startingSolution, -1);
        System.out.println("initial="+startingSolution.totalEnergy+", wakelossfree="+1d/wakelossFreeFitness);
//        System.out.println("initial="+1d/fitfun.valueOf(startingPoint, true)+", wakelossfree="+1d/wakelossFreeFitness);
        
//        double[] energyPerTurbine = fitfun.energyPerTurbine(startingPoint, false);
//        double[] energyPerTurbine = fitfun.energyPerTurbine(startingPoint, printEveryEvaluation, null, -1);
        double[] energyPerTurbine = startingSolution.energyPerTurbine;

        WakeModel.printSolution(startingPoint, energyPerTurbine);
        
        SolutionSet population;

	    int evaluations = 0;

		//Fill population.
	    population = new SolutionSet(populationSize);
	    
	    for (int i = 0; i < populationSize; i++){
	    	Solution newSolution = new Solution(problem);
	    	problem.evaluate(newSolution);
                newSolution.setEvaluation(evaluations);
	    	evaluations++;
	    	population.add(newSolution);
	    }
	    
	    //Rank population.
	    population = rankPopulation(population);
	    
        Solution bestSoFar = population.get(0);
	    int bestSoFarEvaluation = evaluations;
	    
	    //Display the best solution.
        displayCoordinates(population.get(0));
        
        if (pauseForGui){
        	System.out.println(1/population.get(0).getObjectives()[0]);
        	System.out.println(population.get(0).totalEnergy);
        	System.out.println("Pause here please.");
        }
        

        int timesIsBetter = 0;
        int timesIsBetterCounter = 0;
        
        SolutionSet children = new SolutionSet(population.size());
        
	    //Perform generational evaluations
        while (evaluations < maxEvaluations){
//                SolutionSet children = new SolutionSet(population.size());
                SolutionSet union = new SolutionSet(population.size() + children.size());

            	
            	
            	if (evaluations>0 && (evaluations-bestSoFar.getEvaluation())>10000 ) {
                    System.out.println("break: no improvement for evals="+10000);
                    break;
                }

                // two lines of code for more frequent output of the EA
                if (printSeveralOutputsPerDSRun) {
//                    if (evaluations % 5 ==1 ) System.out.print("["+evaluations+":"+1d/bestSoFar.getObjectives()[0]+"]");
//                    if (evaluations % 25 ==0 && evaluations>0) System.out.println();
                    if (evaluations % 1000 ==0 && evaluations>0) {
                        System.out.println("best placement found so far");
//                        energyPerTurbine = fitfun.energyPerTurbine(bestSoFar.getDecisionVariableAsDoubleArray(), false);
//                        energyPerTurbine = fitfun.energyPerTurbine(bestSoFar.getDecisionVariableAsDoubleArray(), false, null, -1); // slow evaluation
                        energyPerTurbine = bestSoFar.energyPerTurbine; 
                        WakeModelFaster.printSolution(bestSoFar, energyPerTurbine);
                    }
                }
                if (printSeveralOutputsPerDSRun && evaluations > population.size() && evaluations%100==0) {
                    System.out.println("[generation:" + evaluations +
                            ", bestSoFar: "+(1.0/bestSoFar.getObjectives()[0])+
                            ", distanceToWakelossFree:" + ((1.0/bestSoFar.getObjectives()[0]) - (1.0/wakelossFreeFitness)) +
                            ", evaluationNumber:" + bestSoFar.getEvaluation() +
                            ", wakeFreeRatio:" + (1.0/bestSoFar.getObjectives()[0]) / (1.0/wakelossFreeFitness) +
                            "]");
                    Solution bestRecent = rankPopulation(children).get(0);
                    System.out.println("[generation:" + evaluations +
                            ", bestRecent:"+(1.0/bestRecent.getObjectives()[0])+
                            ", bestRecent(1/x):"+(bestRecent.getObjectives()[0])+
                            ", distanceToWakelossFree:" + ((1.0/bestRecent.getObjectives()[0]) - (1.0/wakelossFreeFitness)) +
                            ", evaluationNumber:" + bestRecent.getEvaluation() +
                            ", wakeFreeRatio:" + (1.0/bestSoFar.getObjectives()[0]) / (1.0/wakelossFreeFitness) +
                            "]" );
                }
                children = new SolutionSet(population.size());
            	
            	
            	

                if (printEveryEvaluation) {
                    System.out.println("eval="+evaluations+", current="+1d/population.get(0).getObjectives()[0] + ", best found at eval=" + bestSoFarEvaluation);
                }


                for (int i = 0; i < populationSize; i++){
                        Solution newSolution = (Solution)DeepCopy.copy(population.get(i));
//                        Solution newSolution = new Solution(population.get(i));
                        newSolution = (Solution)mutationOperator.execute(newSolution);

//                        problem.evaluate(newSolution);                        // slow evaluation
                        fitfun.updateFast(printEveryEvaluation, newSolution);     // faster evaluation
                        newSolution.setObjective(0, 1d/newSolution.totalEnergy);
                        
                        
                        
                        newSolution.setEvaluation(evaluations);
                        evaluations++;
                        
                        //Adaptive node mutation alteration
                        boolean[] mutatedNodes = mutationOperator.getMutatedNodes();
                        boolean better = newSolution.getObjectives()[0] < population.get(i).getObjectives()[0];
                        for (int node = 0; node < mutatedNodes.length; node++){
                        	if (mutatedNodes[node]){
                        		newSolution.update(node,better);
                                        mutatedNodes[node]=false;
                        	}
                        }
                        	
                        
                        if (printEveryEvaluation) {
                            System.out.println(" new="+evaluations+", current="+1d/newSolution.getObjectives()[0] + ", best found at eval=" + bestSoFarEvaluation);
                        }

                        children.add(newSolution);
                }


                //Rank the combined parent and child populations, and (deterministically) take the fittest as the new population.
                union = children.union(population);
                union = rankPopulation(union);
                
                //Simple rank-based selection
                for (int i = 0; i < population.size(); i++){
                        population.set(i, union.get(i));
                }
                
//				//Simulated annealing
//    			for (int i = 0; i < population.size(); i++){
//					double popFitness = population.get(i).getObjectives()[0];
//					double childFitness = children.get(i).getObjectives()[0];
//					double probability = 1;
//					if (childFitness >= popFitness){
//						probability = Math.exp((popFitness - childFitness)/(10/evaluations));
//					}
//					if (probability > random .nextDouble()){
//						population.set(i, children.get(i));
//					}
//				}
//				population = rankPopulation(population);

                

                timesIsBetterCounter++;
                if (population.get(0).getObjectives()[0] < bestSoFar.getObjectives()[0]) {
                	bestSoFarEvaluation = evaluations;
                    bestSoFar = population.get(0);
                    timesIsBetter++;
                }
                
//                VectorBasedMovementMutation v = (VectorBasedMovementMutation)mutationOperator;
//                double mu = v.getDistanceMu(), sigma = v.getDistanceSigma();
//                if (minMu == -1) minMu = mu;
//                if (maxMu == -1) maxMu = mu*2;
//                System.out.println("\n\n" + v.getDistanceMu() + "\n\n");
//                if (timesIsBetterCounter==5) {
//                	if (failures == 40){
//                		failures = 5;
//                		handleFailures = true;
//                	}
//                	if (failures == 0){
//                		handleFailures = false;
//                	}
//                	if (handleFailures){
//                		failures--;
//                		timesIsBetter++;
//                	}
//                	
//                    if (timesIsBetter>=1) {
//                    	mu = mu * 1.2;
//                        if (mu > maxMu) mu = maxMu;
//                        sigma = mu/3;
//                        v.setDistanceMuAndSigma(mu, sigma);
//   
//                    } else{
//                    	mu = mu / 1.2;
//                    	if (mu < minMu) mu = minMu;
//                        sigma = mu/3;
//                        v.setDistanceMuAndSigma(mu, sigma);
//                        failures++;
//                    }
//                    timesIsBetter = 0;
//                    timesIsBetterCounter = 0;
//                }
//                System.out.println("timesIsBetter="+timesIsBetter+" timesIsBetterCounter="+timesIsBetterCounter+" mu="+mu+" sigma="+sigma);

//                if (evaluations%10!=0 && evaluations<1000)population.set(0,children.get(0));

            //Display the best solution every 100 generations.
                if (evaluations % 100 == 0){
                        displayCoordinates(bestSoFar);
//                        displayCoordinates(population.get(0));
                }
                if (pauseForGui && evaluations % 5000 == 0){
                	System.out.println(1/population.get(0).getObjectives()[0]);
                	System.out.println(population.get(0).totalEnergy);
                	System.out.println("Pause here please.");
                }

                if (debugPrint) System.out.println();
            }

            displayCoordinates(population.get(0));

            
            
            if (true) {
                System.out.println("fast evaluated: "+1d/population.get(0).getObjectives()[0]);
                
                fitfun.valueOf(printEveryEvaluation, population.get(0), -1);
                System.out.println("norm evaluated: "+1d/population.get(0).getObjectives()[0]);
            }
            
            
            if (printSeveralOutputsPerDSRun) System.out.println("["+(evaluations-1)+":"+1d/bestSoFar.getObjectives()[0]+"]");
            
            return population;
	}
        
        
        
        
	
	static double minMu = -1;
	static double maxMu = -1;
	static double currentMu = -1;
	static int failures = 0;
	static boolean handleFailures = false;
		
	/**
	 * Return a fitness-ordered version of the supplied population.
	 * 
	 * @param population  The population of solutions to be ordered.
	 * @return  The fitness-ordered version of the supplied population.
	 */
	private SolutionSet rankPopulation(SolutionSet population){
            boolean debugPrint = !true;

            if (debugPrint) System.out.println("rankPopulation/1");
		
            //Create map and add all solutions.
            HashMap<Integer, Double> unsortedMap = new HashMap<Integer, Double>();
            for (int i = 0; i < population.size(); i++){
                double objective = population.get(i).getObjectives()[0];
                unsortedMap.put(i, objective);
                if (debugPrint) System.out.println(" i"+i+"="+objective);
            }

            //Create map with solutions ordered by fitness.
            TreeMap<Integer, Double> sortedMap = new TreeMap<Integer, Double>(new ValueComparator(unsortedMap));
            sortedMap.putAll(unsortedMap);

            Integer sortedKeys[] = new Integer[population.size()];
            sortedMap.keySet().toArray(sortedKeys);

            //Create new population and add ordered solutions.
            SolutionSet newPopulation = new SolutionSet(population.size());
            for (int i = 0; i < sortedKeys.length; i++){
                    newPopulation.add(population.get(sortedKeys[i]));
            }

            return newPopulation;
	}
	
	/**
	 * Inner comparator class to allow map value-based ordering.
	 *  Orders value based on supplied keys.
	 */
	private static class ValueComparator<Key, Value> implements Comparator<Value>{

		private Map<Key,Value> data;
		
		/**
		 * Instantiate the comparator.
		 * Requires a Map with the same elements as the Map being sorted.
		 * @param data  The Map with the same elements as the Map being sorted.
		 */
		public ValueComparator(Map<Key,Value> data){
			this.data = data;
		}
		
		/**
		 * Compares the order of two keys based on their values.
		 */
		public int compare(Object arg0, Object arg1) {
			Value d0 = (Value) data.get(arg0);
			Value d1 = (Value) data.get(arg1);
			int returnThis = ((Comparable<Value>)d0).compareTo(d1);
			//Elements that are seen as equal are discarded,
			//  thus we force equality to be impossible.
			if (returnThis == 0){
				returnThis++;
			}
			return returnThis;
		}
	}
	
	/**
	 * Output the selected solution to a GUI.
	 * 
	 * @param solution  The solution selected to display.
	 */
	private void displayCoordinates(Solution solution) {
		if (displayGUI) gui.updateSolution(solution);
	}
}
