package directSpread.utility;

import java.io.Serializable;

public class Solution implements Serializable{

    private Variable[] vars;
    private double[] objectiveValues;
    private int evaluation;
    private double[] distanceMus;
    private double[] distanceSigmas;
    private double[] directionSigmas;
    private Problem problem;

    // the following fields are used by the faster fitness fitness evaluation
    public double[] energyPerTurbine;
    public double totalEnergy;
    public boolean[] turbineMoved;
    // store for each wind direction theta a table of velocity deficits 
    public double[][][] thetaVeldefijMatrix;
    // store for each wind direction theta
    public double[][] thetaVeldefiVector;
    
    /**
     * Get the problem being used in this solution.
     * @return  The problem being used in this solution.
     */
    public Problem getProblem() {
        return problem;
    }

    /**
     * Constructor, sets up variables for the solution.
     * 
     * @param problem  The problem being used in this solution.
     */
    public Solution(Problem problem) {
        this.problem = problem;
        vars = problem.createVariables();
        objectiveValues = new double[]{Double.MAX_VALUE};

        distanceMus = problem.getInitialDistanceMus();
        distanceSigmas = problem.getInitialDistanceSigmas();
        directionSigmas = problem.getInitialDirectionSigmas();
        
        
        
//        vars.length
//        this.coordinates = coordinates;
        this.energyPerTurbine = new double[vars.length/2];
        this.totalEnergy = 0;
        
        this.turbineMoved = new boolean[vars.length/2];
        for (int i = 0; i<this.turbineMoved.length; i++) {
            this.turbineMoved[i] = false;
        }
        
        
        this.thetaVeldefijMatrix = new double[24][vars.length/2][vars.length/2];
        this.thetaVeldefiVector = new double[24][vars.length/2];
    }
    
    
    public Solution(double[] d) {
//        this.problem = problem;
//        vars = problem.createVariables();
//        objectiveValues = new double[]{Double.MAX_VALUE};
//
//        distanceMus = problem.getInitialDistanceMus();
//        distanceSigmas = problem.getInitialDistanceSigmas();
//        directionSigmas = problem.getInitialDirectionSigmas();
        this.setDecisionVariablesAsDoubleArray(d);
        
        
//        vars.length
//        this.coordinates = coordinates;
        this.energyPerTurbine = new double[vars.length/2];
        this.totalEnergy = 0;
        
        this.turbineMoved = new boolean[vars.length/2];
        for (int i = 0; i<this.turbineMoved.length; i++) {
            this.turbineMoved[i] = false;
        }
        
        
        this.thetaVeldefijMatrix = new double[24][vars.length/2][vars.length/2];
        this.thetaVeldefiVector = new double[24][vars.length/2];
    }
    
    
    
    public void updateLocation(int turbine, double x, double y) {
        this.vars[turbine*2].setValue(x);
        this.vars[turbine*2+1].setValue(y);
        this.turbineMoved[turbine] = true;
    }
    
    

//    /**
//     * Copy constructor. Copies every variable of the provided solution.
//     * 
//     * @param solution  The solution to copy.
//     */
//    public Solution(Solution solution){
//        
//        Variable[] sv = solution.getDecisionVariables();
//        vars = new Variable[sv.length];
//        for (int i = 0; i < vars.length; i++){
//                vars[i] = new Variable(sv[i].getValue());
//        }
//
//        objectiveValues = new double[solution.getObjectives().length];
//        for (int i = 0; i<objectiveValues.length; i++) {
//            objectiveValues[i] = solution.getObjectives()[i];
//        }
//        
//        int numNodes = vars.length/2;
//        distanceMus = new double[numNodes];
//        distanceSigmas = new double[numNodes];
//        directionSigmas = new double[numNodes];
//        
//        for (int i = 0; i < numNodes; i++){
//        	distanceMus[i] = solution.getDistanceMu(i);
//        	distanceSigmas[i] = solution.getDistanceSigma(i);
//        	directionSigmas[i] = solution.getDirectionSigma(i);
//        }
//        
//        problem = solution.getProblem();
//    }
    /**
     * Return the decision variables of the solution.
     * @return  The decision variables of the solution.
     */
    public Variable[] getDecisionVariables() {
        return vars;
    }

    public double[] getDecisionVariableAsDoubleArray() {
        double[] coords = new double[vars.length];
        for (int i = 0; i < vars.length; i++) {
            coords[i] = vars[i].getValue();
        }
        return coords;
    }

    /**
     * Set the decision variables of the solution.
     * @param variables  The decision variables of the solution.
     */
    public void setDecisionVariables(Variable[] variables) {
        vars = variables;
    }
    
    public void setDecisionVariablesAsDoubleArray(double[] d){
        vars = new Variable[d.length];
        for (int i=0; i<d.length; i++) {
            vars[i] = new Variable(d[i]);
        }
    }

    /**
     * Set an objective of the solution.
     * @param i  The objective's index.
     * @param d  The objective's value.
     */
    public void setObjective(int i, double d) {
        objectiveValues[i] = d;
    }

    public void setEvaluation(int e) {
        this.evaluation = e;
    }

    public int getEvaluation() {
        return this.evaluation;
    }

    public double[] getObjectives() {
        return objectiveValues;
    }

    public double getDistanceMu(int i) {
        if (i < 0 || i > distanceMus.length) {
            throw new ArrayIndexOutOfBoundsException(i);
        }
        return distanceMus[i];
    }

    public double getDistanceSigma(int i) {
        if (i < 0 || i > distanceSigmas.length) {
            throw new ArrayIndexOutOfBoundsException(i);
        }
        return distanceSigmas[i];
    }

    public double getDirectionSigma(int i) {
        if (i < 0 || i > directionSigmas.length) {
            throw new ArrayIndexOutOfBoundsException(i);
        }
        return directionSigmas[i];
    }

    /**
     * Update a node's distance modifier properties based on the success or
     *  failure of mutation upon the node.
     * 
     * @param node  The node to update.
     * @param better  Whether the node's new location is better.
     */
    public void update(int node, boolean better) {
        boolean debugPrint = false;
        if (debugPrint) {
            System.out.println("Node " + node + ", mu: " + distanceMus[node] + ", sigma: " + distanceSigmas[node]);
        }
        if (node < 0 || node > distanceMus.length) {
            throw new ArrayIndexOutOfBoundsException(node);
        }
        if (better) {
            distanceMus[node] = Math.min(distanceMus[node] * problem.getDistanceMuIncrease(), problem.getDistanceMuMax());
            distanceSigmas[node] = Math.min(distanceSigmas[node] * problem.getDistanceSigmaIncrease(), problem.getDistanceSigmaMax());
        } else {
            distanceMus[node] = Math.max(distanceMus[node] / problem.getDistanceMuDecrease(), problem.getDistanceMuMin());
            distanceSigmas[node] = Math.max(distanceSigmas[node] / problem.getDistanceSigmaDecrease(), problem.getDistanceSigmaMin());
        }
        if (debugPrint) {
            System.out.println("Node " + node + " is better: " + better + ", mu: " + distanceMus[node] + ", sigma: " + distanceSigmas[node]);
        }
    }
}